/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.connector.aws.util;

import org.apache.flink.annotation.Internal;
import org.apache.flink.connector.base.sink.throwable.FatalExceptionClassifier;

import software.amazon.awssdk.core.exception.SdkClientException;
import software.amazon.awssdk.services.sts.model.StsException;

/** Class containing set of {@link FatalExceptionClassifier} for AWS credential failures. */
@Internal
public class AWSCredentialFatalExceptionClassifiers {
    public static FatalExceptionClassifier getInvalidCredentialsExceptionClassifier() {
        return FatalExceptionClassifier.withRootCauseOfType(
                StsException.class,
                err ->
                        new AWSAuthenticationException(
                                "Encountered non-recoverable exception relating to the provided credentials.",
                                err));
    }

    public static FatalExceptionClassifier getSdkClientMisconfiguredExceptionClassifier() {
        return FatalExceptionClassifier.withRootCauseOfType(
                SdkClientException.class, err -> (Exception) err);
    }
}
